// -*- mode: c++; mode: fold -*-
// Description                              /*{{{*/
// $Id: dudf.cc,v 1.2 2002/07/25 18:07:18 prrt Exp $
/* ######################################################################

 DUDF - Distribution Upgradeability Description Format

 Implementation of DUDF
 It is responsible for generating the DUDF file as defined in MANCOOSI
 project.
 DUDF contais one problem instance.
 Inspired on GlobalError class written by Jason Gunthorpe.

 ##################################################################### */
/*}}}*/
// Include Files                            /*{{{*/
#include "dudf.h"
#include <fstream>
#include <iostream>
#include <sstream>
#include <cerrno>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/param.h>

#include <uuid++.hh>
#include <json/json.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

//For rpm DB Dump
#define _RPMEVR_INTERNAL 1
#include <rpmtypes.h>
#include <rpmrc.h>
#include <rpmts.h>
#include <rpmds.h>
#include <rpmdb.h>
#include <rpmmacro.h>
#include <rpmio.h>
#include <rpmversion.h>


const std::string SafeGetCWD() {
    // Stash the current dir.
    char S[MAXPATHLEN];
    S[0] = 0;
    if (getcwd(S, sizeof(S) - 2) == 0)
        return "/";
    unsigned int Len = strlen(S);
    S[Len] = '/';
    S[Len + 1] = 0;
    return S;
}

/*
 * Retrieve and parse official repos from config files
 */
void getOfficialRepos(std::vector<std::string> &repos) {
    //SL:std::string repos_option = _config->Find("APT::Dudf::OfficialRepos", "");
    std::string repos_option = "";
    std::string delimiters = ";";

    // Skip delimiters at beginning.
    std::string::size_type lastPos = repos_option.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    std::string::size_type pos = repos_option.find_first_of(delimiters, lastPos);

    while (std::string::npos != pos || std::string::npos != lastPos) {
        // Found a token, add it to the std::vector.
        repos.push_back(repos_option.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
        lastPos = repos_option.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = repos_option.find_first_of(delimiters, lastPos);
    }
}

const std::string md5OfFile(const std::string &Filename) {

    const unsigned char *digest = NULL;
    struct stat sb;
    size_t digestlen = 0;
    char buf[8*BUFSIZ];
    size_t nb;

    if(Stat(Filename.c_str(), &sb) >= 0) {
	DIGEST_CTX ctx = rpmDigestInit(PGPHASHALGO_MD5, RPMDIGEST_NONE);
	FD_t fd = Fopen(Filename.c_str(), "r");
	int xx;

	while ((nb = Fread(buf, sizeof(buf[0]), sizeof(buf), fd)) > 0) {
	    xx = rpmDigestUpdate(ctx, buf, nb);
	}
	xx = Fclose(fd);
	xx = rpmDigestFinal(ctx, &digest, &digestlen, 0);
    }
    if (!digestlen)
	return "";
    const std::string hexStr(pgpHexStr(digest, digestlen));
    _free(digest);
    return hexStr;
}

bool isDepFlag(uint64_t value) {
    return (value & (RPMSENSE_LESS | RPMSENSE_GREATER | RPMSENSE_EQUAL));
}

const Json::Value getTagArray(Header hdr, rpmTag tagname)
{
	HE_t tagdata = static_cast<HE_t>(memset(alloca(sizeof(*tagdata)), 0, sizeof(*tagdata)));
	Json::Value arr(Json::arrayValue);

	tagdata->tag = tagname;
	if (headerGet(hdr, tagdata, 0))
	{
	    for (tagdata->ix = 0; tagdata->ix < static_cast<int>(tagdata->c); tagdata->ix++) {
		arr.append(Json::Value(tagdata->p.argv[tagdata->ix]));
	    }
	}
	_free(tagdata->p.ptr);
	return arr;

}

const Json::Value getTagInteger(Header hdr, rpmTag tag)
{
    HE_t tagdata = static_cast<HE_t>(memset(alloca(sizeof(*tagdata)), 0, sizeof(*tagdata)));
    Json::Value val;

    tagdata->tag = tag;
    if (headerGet(hdr, tagdata, 0))
    {
	uint64_t num = 0;
	int ix = (tagdata->ix >= 0 ? tagdata->ix : 0);

	switch (tagdata->t) {
	    case RPM_UINT64_TYPE:
		num = tagdata->p.ui64p[ix];
		break;
	    case RPM_UINT32_TYPE:
		num = tagdata->p.ui32p[ix];
		break;
	    case RPM_UINT16_TYPE:
		num = tagdata->p.ui16p[ix];
		break;
	    case RPM_UINT8_TYPE:
		num = tagdata->p.ui8p[ix];
		break;
	    default:
		break;
	}

	val = Json::Value(Json::UInt(num)); 
	_free(tagdata->p.ptr);
    }
    return val;
}

const Json::Value getTagString(Header hdr, rpmTag tag)
{
    HE_t tagdata = static_cast<HE_t>(memset(alloca(sizeof(*tagdata)), 0, sizeof(*tagdata)));
    Json::Value val;
    tagdata->tag = tag;

    if (headerGet(hdr, tagdata, 0)) {
	const char *str = NULL;
	if (tagdata->t == RPM_STRING_TYPE)
	    str = tagdata->p.str;
	else if(tagdata->t == RPM_STRING_ARRAY_TYPE || tagdata->t == RPM_I18NSTRING_TYPE)
	    str = tagdata->p.argv[tagdata->ix >= 0 ? tagdata->ix : 0];

	val = Json::Value(str);
	_free(tagdata->p.ptr);
    }
    return val;

}

/*
 * Generic getter for versioned dependency style info
 * It should work for requires, provides, conflicts and obsoletes
 * TODO: Maybe encode the requires as [name, flag, version] arrays as suggested by Pietro
 * see https://gforge.info.ucl.ac.be/tracker/?func=detail&aid=262&group_id=35
 *
 */
const Json::Value getTagRequires(Header hdr, rpmTag tag, rpmTag flag_tag,
	rpmTag version_tag)
{
    HE_t tagdata1 = static_cast<HE_t>(memset(alloca(sizeof(*tagdata1)), 0, sizeof(*tagdata1)));
    HE_t tagdata2 = static_cast<HE_t>(memset(alloca(sizeof(*tagdata2)), 0, sizeof(*tagdata2)));
    HE_t tagdata3 = static_cast<HE_t>(memset(alloca(sizeof(*tagdata3)), 0, sizeof(*tagdata3)));
    Json::Value arr(Json::arrayValue);
    std::set<std::string, StringComparator> depSet;

    tagdata3->tag = version_tag;
    headerGet(hdr, tagdata3, 0);
    tagdata1->tag = tag;
    headerGet(hdr, tagdata1, 0);

    tagdata2->tag = flag_tag;
    if (headerGet(hdr, tagdata2, 0)) {
	uint64_t depflag=0;
	for (tagdata2->ix = 0; tagdata2->ix < static_cast<int>(tagdata2->c); tagdata2->ix++) {
	    switch (tagdata2->t) {
		case RPM_UINT64_TYPE:
		    depflag = tagdata2->p.ui64p[tagdata2->ix];
		    break;
		case RPM_UINT32_TYPE:
		    depflag = tagdata2->p.ui32p[tagdata2->ix];
		    break;
		case RPM_UINT16_TYPE:
		    depflag = tagdata2->p.ui16p[tagdata2->ix];
		    break;
		case RPM_UINT8_TYPE:
		    depflag = tagdata2->p.ui8p[tagdata2->ix];
		    break;
		default:
		    break;
	    }
	    tagdata1->ix++;	//Requirename 
	    tagdata3->ix++; //Requireversion
	    std::string dep(tagdata1->p.argv[tagdata1->ix]);
	    if (depflag & RPMSENSE_RPMLIB) //Starts with "rpmlib("
		continue;

	    if(isDepFlag(depflag))
	    {
		static const char *Fstr[] = { "?0","<",">","?3","=","<=",">=","?7" };
		uint32_t Fx = ((depflag >> 1) & 0x7);
		dep += " ";
		dep += Fstr[Fx];
		dep += " ";
		// version
		dep += tagdata3->p.argv[tagdata3->ix];
	    }
	    depSet.insert(dep);

	}

	//Using depSet assures no duplicates in a set of Deps
	for (std::set<std::string>::const_iterator it=depSet.begin(); it != depSet.end(); ++it)
	    arr.append(Json::Value(*it));
    }
    _free(tagdata1->p.ptr);
    _free(tagdata2->p.ptr);
    _free(tagdata3->p.ptr);

    return arr;

}

// Dudf Object                          /*{{{*/
GlobalDudf *_GetDudfObj() {
    static GlobalDudf *Obj = new GlobalDudf;
    return Obj;
}
/*}}}*/

// GlobalDudf::GlobalDudf - Constructor             /*{{{*/
// ---------------------------------------------------------------------
/* */
GlobalDudf::GlobalDudf() :
    /* Default to success with no force dudf */
    urpmiRequestStatus(true), urpmiForceDudf(false),
    urpmiRequest(), PtrPkgInfo(), baseSystemDeps(),
    /* No virtual package selection has been issued yet! */    
    virtualPackageSelection(false),
    virtualPackages(), tmpVirtualPackage(),
    /* Initializes Meta Installer with empty values: user will set it */
    metaInstallerName(), metaInstallerVersion(),
    /* But define the rpm version/name! Symbols are exported from librpm */
    rpmName(rpmNAME), rpmVersion(rpmEVR),
    FilenameFingerprint(80, 0), Timestamp(), DudfErrorMsg(),
    DudfLogMsg(), DudfInfoElement(), dudfDoc(NULL), rpmStateDir(), urpmiConfig(),
    urpmiMediaCfgDir(), urpmiSkiplist(), urpmiInstlist(), urpmiPreferList(),
    urpmiPreferVendorlist(), urpmiPrivateNetrc(), urpmiStateDir(), urpmiCacheDir(),
    urpmiRoot(), urpmiEnvStatus(false)
	{
    time_t Date;
    struct tm * timeinfo;
    char *oldLCTIME;

    oldLCTIME = setlocale(LC_TIME, "C");
    if (!oldLCTIME) {
	std::cerr << "Warning: setlocale() returned NULL. Unable to set LC_TIME, DUDF might be invalid!" << std::endl;
    }

    time(&Date);
    timeinfo = localtime(&Date);

    strftime(const_cast<char *>(FilenameFingerprint.data()), 80, "%Y%m%d%H%M%S", timeinfo);
    strftime(const_cast<char *>(Timestamp.data()), 80, "%a, %d %b %Y %H:%M:%S GMT", timeinfo);


    //rpmdata = RPMPackageData::Singleton();

    oldLCTIME = setlocale(LC_TIME, oldLCTIME);
    if (!oldLCTIME) {
	std::cerr << "Warning: setlocale() returned NULL. Unable to reset LC_TIME." << std::endl;
    }

    xmlInitParser();
}

/**
 * Dump the RPM installation status into a JSON object to be embedded in the DUDF
 * TODO: document the json format used
 */
std::string GlobalDudf::dumpRPMDb() {
    std::stringstream rpmDbDump;
    //TODO: We should be reusing the RPM db connection that apt already has, maybe
    // through RPMDBHandler class
    rpmts ts = rpmtsCreate();
    Header hdr;
    if (rpmReadConfigFiles(NULL, NULL) == -1) {
	std::cerr << "Error while reading config files." << std::endl;
	return "";
    }

    if (!isUrpmiEnvSet()) {
	std::cerr << "URPMI environment wasn't set. Please set all URPMI paths before getting a DUDF." << std::endl;
	return "";
    }

    rpmtsSetRootDir(ts, getUrpmiRoot().c_str());
    Json::Value array_of_pkgs(Json::arrayValue);

    if (rpmtsGetRdb(ts) == NULL) {
        int rc = rpmtsOpenDB(ts, O_RDONLY);

	if (rc != 0) {
		std::cerr << "Error while opening RPM database. Return code: " << rc << std::endl;
		return "";
	}
    }

    constructBaseSystemDependencies(ts);
    rpmmi Iter = rpmtsInitIterator(ts, RPMTAG_NAME, NULL, 0);
    while ((hdr = rpmmiNext(Iter)) != NULL) {
        const Json::Value &name = getTagString(hdr, RPMTAG_NAME);
        //Skipping RPM pseudo-packages for GPG pubkeys
        if (strcmp(name.asCString(), "gpg-pubkey") == 0)
            continue;

        Json::Value pkg(Json::arrayValue);

        pkg.append(name);
        pkg.append(getTagInteger(hdr, RPMTAG_EPOCH));
        pkg.append(getTagString(hdr, RPMTAG_VERSION));
        pkg.append(getTagString(hdr, RPMTAG_RELEASE));
        pkg.append(getTagRequires(hdr, RPMTAG_REQUIRENAME,
                RPMTAG_REQUIREFLAGS, RPMTAG_REQUIREVERSION));
        pkg.append(getTagRequires(hdr, RPMTAG_PROVIDENAME,
                RPMTAG_PROVIDEFLAGS, RPMTAG_PROVIDEVERSION));
        pkg.append(getTagRequires(hdr, RPMTAG_CONFLICTNAME,
                RPMTAG_CONFLICTFLAGS, RPMTAG_CONFLICTVERSION));
        pkg.append(getTagRequires(hdr, RPMTAG_OBSOLETENAME,
                RPMTAG_OBSOLETEFLAGS, RPMTAG_OBSOLETEVERSION));
        pkg.append(getTagInteger(hdr, RPMTAG_SIZE));
        const std::string &tmp = name.asString();
        pkg.append(Json::Value(isEssentialPackage(tmp)));

        array_of_pkgs.append(pkg);
    }

    headerFree(hdr);
    rpmmiFree(Iter);
    rpmtsFree(ts);
    rpmFreeMacros(NULL);
    rpmFreeRpmrc();

    //SL: bool pretty_json = _config->FindB("APT::Dudf::PrettyJson", false);
	bool pretty_json = false;

    if (pretty_json) {
        Json::StyledWriter pretty_writer;
        rpmDbDump << pretty_writer.write(array_of_pkgs) << std::endl;
    } else {
        Json::FastWriter fast;
        rpmDbDump << fast.write(array_of_pkgs) << std::endl;
    }


    return rpmDbDump.str();

}

std::string GlobalDudf::generateUUID(void) {
    uuid generator;
    generator.make(UUID_MAKE_V4);

    char *tmp = generator.string();
    std::string retval(tmp);
    free(tmp);

    return retval;
}

void GlobalDudf::constructBaseSystemDependencies(rpmts ts) {
	const char *dep = NULL, *name = NULL;
	Header hdr;
	HE_t tagdata = static_cast<HE_t>(memset(alloca(sizeof(*tagdata)), 0, sizeof(*tagdata)));
	rpmmi Iter = rpmtsInitIterator(ts, RPMTAG_NAME, NULL, 0);

	tagdata->tag = RPMTAG_NAME;

	while ((hdr = rpmmiNext(Iter)) != NULL) {
		if (headerGet(hdr, tagdata, 0)) {
			name = tagdata->p.str;
			if (strstr(name, "basesystem") != NULL) {
				tagdata->tag = RPMTAG_REQUIRENAME;
				if (headerGet(hdr, tagdata, 0)) {
					for(tagdata->ix = 0; tagdata->ix < static_cast<int>(tagdata->c); tagdata->ix++) {
						dep = tagdata->p.argv[tagdata->ix];
						baseSystemDeps.insert(std::string(dep));
						_free(tagdata->p.ptr);
					}
				}
			}
		}
	}

	_free(name);
	_free(tagdata->p.ptr);
	rpmmiFree(Iter);
}

bool GlobalDudf::isEssentialPackage(const std::string& name) {
	return (baseSystemDeps.find(name) != baseSystemDeps.end());
}


//  GlobalDudf::GetDUDFTree - Get the xmlDocPtr     /*{{{*/
// ---------------------------------------------------------------------
/* */
xmlDocPtr GlobalDudf::GetDUDFTree(void) {
	return dudfDoc;
}

//  GlobalDudf::DumpToFile- Dump DUDF into XML file     /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::DumpToFile(void) {
	return DumpToFile("");
}
/*}}}*/

//  GlobalDudf::DumpToFile- Dump DUDF into XML file     /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::DumpToFile(std::string fname = "") {

	std::string dudfFilename;
	if (fname.empty()) {
		std::string DudfDirName = getUrpmiCacheDir() + "/dudf/";
		dudfFilename = DudfDirName + "/dudf_" + FilenameFingerprint + ".xml";
	} else {
		dudfFilename = fname;
	}

	if (GlobalDudf::TryCreateDudfDocument())
		xmlSaveFormatFileEnc( dudfFilename.c_str() , dudfDoc, "UTF-8", 1);
	else
		std::cerr << "Error while generating DUDF file" << std::endl;

	return true;
}
/*}}}*/

//  GlobalDudf::GetFilenameFingerprint- Accessor for the FilenameFingerprint attribute /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::GetFilenameFingerprint() {
	return FilenameFingerprint;
}
/*}}}*/

//  GlobalDudf::setUrpmiRequestStatus - setter for urpmiRequestStatus /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::setUrpmiRequestStatus(bool value) {
	urpmiRequestStatus = value;
}
/*}}}*/

//  GlobalDudf::getUrpmiRequestStatus - getter for urpmiRequestStatus /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::getUrpmiRequestStatus(void) {
	return urpmiRequestStatus;
}
/*}}}*/

//  GlobalDudf::setUrpmiForceDudf - setter for urpmiForceDudf /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::setUrpmiForceDudf(bool value) {
	urpmiForceDudf = value;
}
/*}}}*/

//  GlobalDudf::getUrpmiForceDudf - getter for urpmiForceDudf /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::getUrpmiForceDudf(void) {
	return urpmiForceDudf;
}
/*}}}*/

//  GlobalDudf::getUrpmiResult - return a std::string for dudf:result attribute /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getUrpmiResult(void) {
	/* URPMI succeeded, don't care about force DUDF */
	if (getUrpmiRequestStatus()) return "success";
	if (!getUrpmiRequestStatus()) return "failure";

	return "failure";
}
/*}}}*/

//  GlobalDudf::setUrpmiRequest - setting the urpmi request /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::setUrpmiRequest(std::string value) {
	urpmiRequest = value;
}
/*}}}*/

//  GlobalDudf::getUrpmiRequest - return the registered urpmi request /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getUrpmiRequest(void) {
	return urpmiRequest;
}
/*}}}*/

//  GlobalDudf::setMetaInstallerName - set the value of metaInstallerName /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::setMetaInstallerName(std::string value) {
	metaInstallerName = value;
}
/*}}}*/

//  GlobalDudf::getMetaInstallerName - return the value of metaInstallerName /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getMetaInstallerName(void) {
	return metaInstallerName;
}
/*}}}*/

//  GlobalDudf::setMetaInstallerVersion - set the value for metaInstallerVersion /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::setMetaInstallerVersion(std::string value) {
	metaInstallerVersion = value;
}
/*}}}*/

//  GlobalDudf::getMetaInstallerVersion - return the value of metaInstallerVersion /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getMetaInstallerVersion(void) {
	return metaInstallerVersion;
}
/*}}}*/

//  GlobalDudf::setRpmName - set the value of rpmName /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::setRpmName(std::string value) {
	rpmName = value;
}
/*}}}*/

//  GlobalDudf::getRpmName - return the value of rpmName /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getRpmName(void) {
	return rpmName;
}
/*}}}*/

//  GlobalDudf::setRpmVersion - set the value of rpmVersion /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::setRpmVersion(std::string value) {
	rpmVersion = value;
}
/*}}}*/

//  GlobalDudf::getRpmVersion - return the value of rpmVersion /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getRpmVersion(void) {
	return rpmVersion;
}
/*}}}*/

//  GlobalDudf::setRpmStateDir - set the value of the RPM state dir (/var/lib/rpm).
//  Doesn't include the urpmi root ! /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::setRpmStateDir(std::string value) {
	rpmStateDir = value;
}
/*}}}*/

//  GlobalDudf::getRpmStateDir - get the value of the RPM state dir.
//  Includes the urpmi root ! /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getRpmStateDir(void) {
	return getUrpmiRoot() + rpmStateDir;
}
/*}}}*/

//  GlobalDudf::getRpmPackageDb - get the path to RPM's Package DB /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getRpmPackageDb(void) {
	return getRpmStateDir() + "/Packages";
}
/*}}}*/

//  GlobalDudf::setUrpmiConfig - set the value of the urpmiConfig variable /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::setUrpmiConfig(std::string value) {
	urpmiConfig = value;
}
/*}}}*/

//  GlobalDudf::getUrpmiConfig - get the value of the urpmiConfig variable /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getUrpmiConfig(void) {
	return urpmiConfig;
}
/*}}}*/

//  GlobalDudf::setUrpmiMediaCfgDir - set the value of urpmiMediaCfgDir /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::setUrpmiMediaCfgDir(std::string value) {
	urpmiMediaCfgDir = value;
}
/*}}}*/

//  GlobalDudf::getUrpmiMediaCfgDir - get the value of urpmiMediaCfgDir /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getUrpmiMediaCfgDir(void) {
	return urpmiMediaCfgDir;
}
/*}}}*/

//  GlobalDudf::setUrpmiSkiplist - set the urpmiSkiplist value /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::setUrpmiSkiplist(std::string value) {
	urpmiSkiplist = value;
}
/*}}}*/

//  GlobalDudf::getUrpmiSkiplist - get the value of urpmiSkiplist /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getUrpmiSkiplist(void) {
	return urpmiSkiplist;
}
/*}}}*/

//  GlobalDudf::setUrpmiInstlist - set the value of urpmiInstlist /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::setUrpmiInstlist(std::string value) {
	urpmiInstlist = value;
}
/*}}}*/

//  GlobalDudf::getUrpmiInstlist - get the value of urpmiInstlist /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getUrpmiInstlist(void) {
	return urpmiInstlist;
}
/*}}}*/

//  GlobalDudf::setUrpmiPreferList - set the value of urpmiPreferList /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::setUrpmiPreferList(std::string value) {
	urpmiPreferList = value;
}
/*}}}*/

//  GlobalDudf::getUrpmiPreferList - get the value of urpmiPreferList /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getUrpmiPreferList(void) {
	return urpmiPreferList;
}
/*}}}*/

//  GlobalDudf::setUrpmiPreferVendorlist - set the value of urpmiPreferVendorlist /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::setUrpmiPreferVendorlist(std::string value) {
	urpmiPreferVendorlist = value;
}
/*}}}*/

//  GlobalDudf::getUrpmiPreferVendorlist - get the value of urpmiPreferVendorlist /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getUrpmiPreferVendorlist(void) {
	return urpmiPreferVendorlist;
}
/*}}}*/

//  GlobalDudf::setUrpmiPrivateNetrc - set the value of urpmiPrivateNetrc /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::setUrpmiPrivateNetrc(std::string value) {
	urpmiPrivateNetrc = value;
}
/*}}}*/

//  GlobalDudf::getUrpmiPrivateNetrc - get the value of urpmiPrivateNetrc /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getUrpmiPrivateNetrc(void) {
	return urpmiPrivateNetrc;
}
/*}}}*/

//  GlobalDudf::setUrpmiStateDir - set the value of urpmiStateDir /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::setUrpmiStateDir(std::string value) {
	urpmiStateDir = value;
}
/*}}}*/

//  GlobalDudf::getUrpmiStateDir - get the value of urpmiStateDir /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getUrpmiStateDir(void) {
	return urpmiStateDir + "/";
}
/*}}}*/

//  GlobalDudf::setUrpmiCacheDir - set the value of urpmiCacheDir /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::setUrpmiCacheDir(std::string value) {
	urpmiCacheDir = value;
}
/*}}}*/

//  GlobalDudf::getUrpmiCacheDir - get the value of urpmiCacheDir /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getUrpmiCacheDir(void) {
	return urpmiCacheDir + "/";
}
/*}}}*/

//  GlobalDudf::setUrpmiRoot - set the value of urpmiRoot /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::setUrpmiRoot(std::string value) {
	urpmiRoot = value;
}
/*}}}*/

//  GlobalDudf::getUrpmiRoot - get the value of urpmiRoot /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getUrpmiRoot(void) {
	return urpmiRoot + "/";
}
/*}}}*/

//  GlobalDudf::urpmiEnvSet - specify that all URPMI paths were set! /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::urpmiEnvSet(void) {
	urpmiEnvStatus = true;
}
/*}}}*/

//  GlobalDudf::isUrpmiEnvSet - check wether the URPMI environment is set /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::isUrpmiEnvSet(void) {
	return urpmiEnvStatus;
}
/*}}}*/

//  GlobalDudf::CreateDudf- Retrieve the XML file in a std::string /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::CreateDudf(void) {
	xmlChar* buf = NULL;
	int length = 0;

	if (GlobalDudf::TryCreateDudfDocument())
		xmlDocDumpFormatMemoryEnc(dudfDoc, &buf, &length, "UTF-8", 1);
	else
		std::cerr << "Error while generating DUDF file" << std::endl;

	std::string xmlFile(const_cast<const char *>(reinterpret_cast<char *>(buf)), static_cast<size_t>(length));
	xmlFree(buf);

	return xmlFile;
}
/*}}}*/

//  GlobalDudf::TryCreateDudfDocument- Create document if needed /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::TryCreateDudfDocument(void) {
	GlobalDudf::CreateDudfDocument();
	return (dudfDoc != NULL);
}
/*}}}*/

//  GlobalDudf::setVirtualPackageNameDependency - register the dependency name /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::setVirtualPackageNameDependency(std::string value) {
	tmpVirtualPackage.virtualPackageNameDependency = value;
}
/*}}}*/

//  GlobalDudf::getVirtualPackageNameDependency - get the dependency name /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getVirtualPackageNameDependency(void) {
	return tmpVirtualPackage.virtualPackageNameDependency;
}
/*}}}*/

//  GlobalDudf::setVirtualPackageNameSelection - register choice of user /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::setVirtualPackageNameSelection(std::string value) {
	tmpVirtualPackage.virtualPackageNameSelection = value;
}
/*}}}*/

//  GlobalDudf::getVirtualPackageNameSelection - retrieve choice of user /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getVirtualPackageNameSelection(void) {
	return tmpVirtualPackage.virtualPackageNameSelection;
}
/*}}}*/

//  GlobalDudf::pushVirtualPackageNameChoice - push a new choice for the virtual package /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::pushVirtualPackageNameChoice(std::string value) {
	tmpVirtualPackage.virtualPackageNameChoices.insert(value);
}
/*}}}*/

//  GlobalDudf::getVirtualPackageNameChoices - return the registered set /*{{{*/
// ---------------------------------------------------------------------
/* */
std::set<std::string> GlobalDudf::getVirtualPackageNameChoices(virtualPackageStorage sto) {
	return sto.virtualPackageNameChoices;
}
/*}}}*/

//  GlobalDudf::getFormattedVirtualPackageNameChoices - Format a one std::string containing the set /*{{{*/
// ---------------------------------------------------------------------
/* */
std::string GlobalDudf::getFormattedVirtualPackageNameChoices(virtualPackageStorage sto) {
	std::string retval;

	for (std::set<std::string>::const_iterator it = sto.virtualPackageNameChoices.begin(); it != sto.virtualPackageNameChoices.end(); ++it) {
		retval += *it + " ";
	}

	return retval;
}
/*}}}*/

//  GlobalDudf::getVectorVirtualPackageNameChoices - Returns a std::vector of the set /*{{{*/
// ---------------------------------------------------------------------
/* */
std::vector<std::string> GlobalDudf::getVectorVirtualPackageNameChoices(virtualPackageStorage sto) {
	std::vector<std::string> res;
	
	for (std::set<std::string>::const_iterator it = sto.virtualPackageNameChoices.begin(); it != sto.virtualPackageNameChoices.end(); ++it) {
		res.push_back(*it);
	}

	return res;
}
/*}}}*/

//  GlobalDudf::getVirtualPackages - get all the virtual packages
//  has been completed. /*{{{*/
// ---------------------------------------------------------------------
/* */
std::vector<virtualPackageStorage> GlobalDudf::getVirtualPackages(void) {
	return virtualPackages;
}
/*}}}*/

//  GlobalDudf::resetTmpVirtualPackage - cleanup the tmpVirtualPackage variable.
//  has been completed. /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::resetTmpVirtualPackage(void) {
	tmpVirtualPackage.virtualPackageNameDependency = "";
	tmpVirtualPackage.virtualPackageNameSelection = "";
	tmpVirtualPackage.virtualPackageNameChoices.clear();
}
/*}}}*/

//  GlobalDudf::commitVirtualPackageSelection - notify that a virtual package selection
//  has been completed. /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::commitVirtualPackageSelection(void) {
	virtualPackages.push_back(tmpVirtualPackage);
	virtualPackageSelection = true;
	resetTmpVirtualPackage();
}
/*}}}*/

//  GlobalDudf::hasVirtualPackageSelection - returns true if a virtual package selection
//  has been commited. /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::hasVirtualPackageSelection(void) {
	return virtualPackageSelection;
}
/*}}}*/

//  GlobalDudf::CreateDudfDocument- Create the XML tree /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::CreateDudfDocument(void) {
	std::string pkgstatus;

	std::string UUID = generateUUID();
	xmlDocPtr doc = NULL;
	xmlNodePtr root_node = NULL, node = NULL, node1 = NULL, node2 = NULL,
		   node3 = NULL, node4 = NULL;
	xmlNsPtr ns = NULL;

	doc = xmlNewDoc(reinterpret_cast<const xmlChar *>("1.0"));
	root_node = xmlNewNode(ns , reinterpret_cast<const xmlChar *>("dudf"));
	ns = xmlNewNs ( root_node, reinterpret_cast<const xmlChar *>("http://www.mancoosi.org/2008/cudf/dudf"), NULL);
	ns = xmlNewNs ( root_node, reinterpret_cast<const xmlChar *>("http://www.mancoosi.org/2008/cudf/dudf"), reinterpret_cast<const xmlChar *>("dudf"));
	/*
	 * Define DUDF version according to the RelaxNG specifications.
	 */
	xmlNewProp(root_node, reinterpret_cast<const xmlChar *>("dudf:version"), reinterpret_cast<const xmlChar *>(DUDF_VERSION));
	xmlDocSetRootElement(doc, root_node);

	/* 
	 * Timestamp
	 */
	/* Time is the same used for fingerprinting files */
	node1 = xmlNewChild(root_node, NULL, reinterpret_cast<const xmlChar *>("timestamp"), reinterpret_cast<const xmlChar *>(Timestamp.c_str()));

	/*
	 * UID
	 */
	node1 = xmlNewChild(root_node, NULL, reinterpret_cast<const xmlChar *>("uid"), reinterpret_cast<const xmlChar *>(UUID.c_str()));

	/* 
	 * The Distribution node and content
	 */

	SetDistro();

	xmlNewChild(root_node, NULL, reinterpret_cast<const xmlChar *>("distribution"),
			reinterpret_cast<const xmlChar *>(DudfInfoElement.Distribution.c_str()));

	/* 
	 * The installer  node. 
	 */
	node = xmlNewChild(root_node, NULL, reinterpret_cast<const xmlChar *>("installer"), NULL);
	node1 = xmlNewChild(node, NULL, reinterpret_cast<const xmlChar *>("name"), reinterpret_cast<const xmlChar *>(getRpmName().c_str()));
	node1 = xmlNewChild(node, NULL, reinterpret_cast<const xmlChar *>("version"), reinterpret_cast<const xmlChar *>(getRpmVersion().c_str()));

	/* 
	 * The meta-installer  node
	 */
	node = xmlNewChild(root_node, NULL, reinterpret_cast<const xmlChar *>("meta-installer"), NULL);
	node1 = xmlNewChild(node, NULL, reinterpret_cast<const xmlChar *>("name"), reinterpret_cast<const xmlChar *>(getMetaInstallerName().c_str()));
	node1 = xmlNewChild(node, NULL, reinterpret_cast<const xmlChar *>("version"), reinterpret_cast<const xmlChar *>(getMetaInstallerVersion().c_str()));

	/* Generate if we have a failure or if the user forced DUDF generation */
	if (!getUrpmiRequestStatus() || getUrpmiForceDudf()) {
		/* 
		 * The problem node
		 */
		node = xmlNewChild(root_node, NULL, reinterpret_cast<const xmlChar *>("problem"), NULL);

		/*  The package status                              */
		node1 = xmlNewChild(node, NULL, reinterpret_cast<const xmlChar *>("package-status"), NULL);
		node2 = xmlNewChild(node1, NULL, reinterpret_cast<const xmlChar *>("installer"), NULL);

		pkgstatus = dumpRPMDb();
		if (pkgstatus == "") {
			std::cerr << "Cannot get RPM database status. Aborting DUDF generation." << std::endl;
			return false;
		}

		const char * status = pkgstatus.c_str();
		node3 = xmlNewChild(node2, NULL, reinterpret_cast<const xmlChar *>("status"), NULL);

		std::string PkgDbFilename = getRpmPackageDb();

		xmlNewProp(node3, reinterpret_cast<const xmlChar *>("filename"), reinterpret_cast<const xmlChar *>(PkgDbFilename.c_str()));
		node4 = xmlNewCDataBlock(doc, reinterpret_cast<const xmlChar *>(status), xmlStrlen(reinterpret_cast<const xmlChar *>(status)));
		xmlAddChild(node3, node4);

		std::vector<std::string> PtrListFiles;
		PackageUniverse(PtrListFiles);

		if (PtrListFiles.empty()) {
			std::cerr << "Error while generating package universe." << std::endl;
			return false;
		}

		/*  The package Universe*/
		node1 = xmlNewChild(node, NULL, reinterpret_cast<const xmlChar *>("package-universe"), NULL);

		std::vector<std::string> official_repos;
		getOfficialRepos(official_repos);

		for (std::vector<std::string>::iterator I = PtrListFiles.begin(); I
				!= PtrListFiles.end(); ++I) {
			std::string NameFileFd = getUrpmiStateDir();
			node2 = xmlNewChild(node1, NULL, reinterpret_cast<const xmlChar *>("package-list"), NULL);

			if (strstr(I->c_str(), "synthesis.hdlist") != NULL)
				xmlNewProp(node2, reinterpret_cast<const xmlChar *>("dudf:format"), reinterpret_cast<const xmlChar *>("synthesis_hdlist"));
			else if (strstr(I->c_str(), "hdlist") != NULL)
				xmlNewProp(node2, reinterpret_cast<const xmlChar *>("dudf:format"), reinterpret_cast<const xmlChar *>("hdlist"));

			xmlNewProp(node2, reinterpret_cast<const xmlChar *>("dudf:filename"), reinterpret_cast<const xmlChar *>((NameFileFd + *I).c_str()));

			bool is_official_repo = official_repos.empty(); //Ignore this check if not properly configured
			for (std::vector<std::string>::const_iterator Repo = official_repos.begin(); Repo
					!= official_repos.end(); ++Repo) {
				//Make sure it's an official repo to find if we should include just a reference for the repo list
				if (I->find(Repo->c_str()) != std::string::npos) {
					is_official_repo = true;
					break;
				}
			}

			/* Not available. Need to parse source lists */
			/* xmlNewProp(node2, reinterpret_cast<const xmlChar *>("format"), reinterpret_cast<const xmlChar *>(ListsFilename.c_str())); */

			//SL: std::string NameFileFd = "/" + _config->FindDir("Dir::State::lists")
			//        + I->c_str();

			if (is_official_repo) {
				/* http://doc4.mandriva.org:8087/synthesis/MD5/download */
				std::string synthesisUrl(SYNTHESIS_URL_SIZE, 0);
				const std::string &md5 = md5OfFile(NameFileFd + *I);
				sprintf(const_cast<char *>(synthesisUrl.data()), SYNTHESIS_URL_FORMAT, md5.c_str());
				xmlNewProp(node2, reinterpret_cast<const xmlChar *>("dudf:url"), reinterpret_cast<const xmlChar *>(synthesisUrl.c_str()));
				node3 = xmlNewChild(node2, NULL, reinterpret_cast<const xmlChar *>("hash"), reinterpret_cast<const xmlChar *>(md5.c_str()));
			}
			else
			{
				int fd_list = open(NameFileFd.c_str(), O_RDONLY);
				if (fd_list != -1)
				{
					struct stat Buf;
					if (fstat(fd_list,&Buf) != 0) {
						return std::cerr << "Error in reading" << std::endl;
					}

					size_t file_size = static_cast<size_t>(Buf.st_size);

					std::string list_content(file_size, 0);

					if (read(fd_list, const_cast<char *>(list_content.data()), file_size) == -1)
						std::cerr << "DUDF, read Unable to read file " << NameFileFd.c_str() << std::endl;

					node4 = xmlNewCDataBlock(doc, reinterpret_cast<const xmlChar *>(list_content.c_str()), xmlStrlen(reinterpret_cast<const xmlChar *>(list_content.c_str())));

					xmlAddChild(node2, node4);

				}

			}

		}

		node1 = xmlNewChild(node, NULL, reinterpret_cast<const xmlChar *>("action"), reinterpret_cast<const xmlChar *>(getUrpmiRequest().c_str()));
		if (hasVirtualPackageSelection()) {
			node2 = xmlNewChild(node, NULL, reinterpret_cast<const xmlChar *>("virtual-packages"), NULL);
				std::vector<virtualPackageStorage> paquets = getVirtualPackages();

				for (std::vector<virtualPackageStorage>::const_iterator pkgIt = paquets.begin(); pkgIt != paquets.end(); ++pkgIt) {
					node3 = xmlNewChild(node2, NULL, reinterpret_cast<const xmlChar *>("virtual-package"), NULL);
					xmlNewChild(node3, NULL, reinterpret_cast<const xmlChar *>("whatprovides"), reinterpret_cast<const xmlChar *>((*pkgIt).virtualPackageNameDependency.c_str()));
					node4 = xmlNewChild(node3, NULL, reinterpret_cast<const xmlChar *>("options"), NULL);

					std::set<std::string> liste = getVirtualPackageNameChoices(*pkgIt);

					for (std::set<std::string>::const_iterator it = liste.begin(); it != liste.end(); ++it) {
						xmlNewChild(node4, NULL, reinterpret_cast<const xmlChar *>("option"), reinterpret_cast<const xmlChar *>((*it).c_str()));
					}

					xmlNewChild(node3, NULL, reinterpret_cast<const xmlChar *>("choice"), reinterpret_cast<const xmlChar *>((*pkgIt).virtualPackageNameSelection.c_str()));
				}
		}

		node1 = xmlNewChild(node, NULL, reinterpret_cast<const xmlChar *>("desiderata"), NULL);
		addUrpmiSkipList(doc, node1);
		addUrpmiInstList(doc, node1);
		addUrpmiPreferList(doc, node1);
		addUrpmiPreferVendorList(doc, node1);
	}

	/*
	 * The outcome node
	 */
	node = xmlNewChild(root_node, NULL, reinterpret_cast<const xmlChar *>("outcome"), NULL);

	/* Generate if we have a success or if the user forced DUDF generation */
	if (getUrpmiRequestStatus() || getUrpmiForceDudf()) {
		/*  The package status                              */
		node1 = xmlNewChild(node, NULL, reinterpret_cast<const xmlChar *>("package-status"), NULL);
		node2 = xmlNewChild(node1, NULL, reinterpret_cast<const xmlChar *>("installer"), NULL);

		pkgstatus = dumpRPMDb();
		if (pkgstatus == "") {
			std::cerr << "Cannot get RPM database status. Aborting DUDF generation." << std::endl;
			return false;
		}
		const char * status = pkgstatus.c_str();

		node3 = xmlNewChild(node2, NULL, reinterpret_cast<const xmlChar *>("status"), NULL);

		std::string PkgDbFilename = getRpmPackageDb();

		xmlNewProp(node3, reinterpret_cast<const xmlChar *>("filename"), reinterpret_cast<const xmlChar *>(PkgDbFilename.c_str()));

		node4 = xmlNewCDataBlock(doc, reinterpret_cast<const xmlChar *>(status), xmlStrlen(reinterpret_cast<const xmlChar *>(status)));

		xmlAddChild(node3, node4);
	}

	/*  Result */
	//Normalize result according to DUDF schema
	xmlNewProp(node, reinterpret_cast<const xmlChar *>("dudf:result"), reinterpret_cast<const xmlChar *>(getUrpmiResult().c_str()));

	/*  Log msg - not printed if non existent */
	if (!DudfLogMsg.empty()) {
		xmlNodePtr cData = xmlNewCDataBlock(doc, reinterpret_cast<const xmlChar *>(DudfLogMsg.c_str()), xmlStrlen(reinterpret_cast<const xmlChar *>(DudfLogMsg.c_str())));
		node1 = xmlNewChild(node, NULL, reinterpret_cast<const xmlChar *>("log"), NULL);
		xmlAddChild(node1, cData);
	}

	/*  Error msg - not printed on sucess */
	if (!DudfErrorMsg.empty()) {
		xmlNodePtr cData = xmlNewCDataBlock(doc, reinterpret_cast<const xmlChar *>(DudfErrorMsg.c_str()), xmlStrlen(reinterpret_cast<const xmlChar *>(DudfErrorMsg.c_str())));
		node1 = xmlNewChild(node, NULL, reinterpret_cast<const xmlChar *>("error"), NULL);
		xmlAddChild(node1, cData);
	}

	if(dudfDoc != NULL)
	    xmlFreeDoc(dudfDoc);
	dudfDoc = doc;
	return true;
}
/*}}}*/

//  GlobalDudf::addUrpmiFile - Add an URPMI file to the XML tree /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::addUrpmiFile(xmlDocPtr doc, xmlNodePtr node, std::string nodeName, std::string filename) {
	std::streamsize length;
	std::string line;
	std::ifstream input;
	xmlNodePtr dataNode, cData;

	input.open(filename.c_str());
	if (!input.is_open()) {
		return false;
	}

	input.seekg(0, std::ios::end);
	length = static_cast<std::streamsize>(input.tellg());
	input.seekg(0, std::ios::beg);

	std::string buffer(static_cast<size_t>(length), 0);
	if (buffer.length() != static_cast<size_t>(length)) {
		input.close();
		return false;
	}

	input.read(const_cast<char *>(buffer.data()), length);
	input.close();

	dataNode = xmlNewChild(node, NULL, reinterpret_cast<const xmlChar *>(nodeName.c_str()), NULL); 
	cData = xmlNewCDataBlock(doc, reinterpret_cast<const xmlChar *>(buffer.c_str()), xmlStrlen(reinterpret_cast<const xmlChar *>(buffer.c_str())));
	xmlNewProp(dataNode, reinterpret_cast<const xmlChar *>("dudf:filename"), reinterpret_cast<const xmlChar *>(filename.c_str()));
	xmlAddChild(dataNode, cData);

	return true;
}
/*}}}*/

//  GlobalDudf::addUrpmiSkipList - Add an URPMI file to the XML tree /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::addUrpmiSkipList(xmlDocPtr doc, xmlNodePtr node) {
	return addUrpmiFile(doc, node, "urpmi_skip", "/etc/urpmi/skip.list");
}
/*}}}*/

//  GlobalDudf::addUrpmiInstList - Add an URPMI file to the XML tree /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::addUrpmiInstList(xmlDocPtr doc, xmlNodePtr node) {
	return addUrpmiFile(doc, node, "urpmi_inst", "/etc/urpmi/inst.list");
}
/*}}}*/

//  GlobalDudf::addUrpmiPreferList - Add an URPMI file to the XML tree /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::addUrpmiPreferList(xmlDocPtr doc, xmlNodePtr node) {
	return addUrpmiFile(doc, node, "urpmi_prefer", "/etc/urpmi/prefer.list");
}
/*}}}*/

//  GlobalDudf::addUrpmiPreferVendorList - Add an URPMI file to the XML tree /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::addUrpmiPreferVendorList(xmlDocPtr doc, xmlNodePtr node) {
	return addUrpmiFile(doc, node, "urpmi_prefer_vendor", "/etc/urpmi/prefer.vendor.list");
}
/*}}}*/


//  GlobalDudf::SetDistro - Set distribution information        /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::SetDistro() {

    std::string File = "/etc/release";

    // See if the file exists
    struct stat Buf;
    if (stat(File.c_str(), &Buf) != 0)
        //SL_error->Error(_("/etc/release does not exist:"));
        std::cerr << "/etc/release does not exist:";

    // Open the stream for reading
    std::ifstream F(File.c_str(), std::ios::in);

    if (!F != 0)
        //_error->Error(_("Could not open /etc/release file:"));
        std::cerr << "Could not open /etc/release file:";

    char Buffer[1024];

    int CurLine = 0;
    while (F.eof() == false) {
        F.getline(Buffer, sizeof(Buffer));
        CurLine++;

        if (F.fail() && !F.eof())
            std::cerr << "Error reading /etc/release: ";

        if (strlen(Buffer) == 0)
            break;

        DudfInfoElement.Distribution = Buffer;

    }

    return true;
}

//  GlobalDudf::PackageUniverse - Include intensional / extensional Package Lists       /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::PackageUniverse(std::vector<std::string> &PtrListFiles) {

    std::string DirLists = getUrpmiStateDir();

    DIR *D = opendir(DirLists.c_str());
    if (D == 0)
        std::cerr << "Unable to read " << DirLists.c_str();

    std::string StartDir = SafeGetCWD();
    if (chdir(DirLists.c_str()) != 0) {
        closedir(D);
        std::cerr << "Unable to change to " << DirLists.c_str();
    }

    for (struct dirent *Dir = readdir(D); Dir != 0; Dir = readdir(D)) {
        // Skip files that aren't lists
	if (strcmp(Dir->d_name, ".") == 0 || strcmp(Dir->d_name, "..") == 0)
	    continue;

	DIR *SD = opendir(Dir->d_name);
	if (SD == NULL) {
		if (errno == ENOTDIR) {
			continue;
		}

		std::cerr << "Unable to read " << Dir->d_name << std::endl;
		break;
	}
	std::string curDir = SafeGetCWD();
	if (chdir(Dir->d_name) != 0) {
		closedir(SD);
		std::cerr << "Unable to change forward to " << Dir->d_name << std::endl;
		break;
	}
	for (struct dirent *Subdir = readdir(SD); Subdir != 0; Subdir = readdir(SD)) {
		if (strcmp(Subdir->d_name, "lock") == 0 || strcmp(Subdir->d_name, "partial")
			== 0 || strcmp(Subdir->d_name, ".") == 0 || strcmp(Subdir->d_name,
			"..") == 0 || strstr(Subdir->d_name, "hdlist") == NULL)
		    continue;

		PtrListFiles.push_back(Dir->d_name + std::string("/") + Subdir->d_name);
	}

        //Args[++argcount] = Dir->d_name;
	closedir(SD);
	if (chdir(curDir.c_str()) != 0) {
		std::cerr << "Unable to change back to " << Dir->d_name;
		break;
	}
    };

    if (chdir(StartDir.c_str()) != 0)
        return false;
    closedir(D);
    return true;

}
/*}}}*/

//  GlobalDudf::AddMsgError - Append to msg error   /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::AddMsgError(std::string MsgErrorPart) {

    DudfErrorMsg += MsgErrorPart;

}
/*}}}*/

//  GlobalDudf::AddMsgLog - Append to msg log   /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::AddMsgLog(std::string MsgLogPart) {

    DudfLogMsg += MsgLogPart;

}
/*}}}*/

//  GlobalDudf::PackagesInfo - Add info about a package /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::PackagesInfo(std::string PkgName, std::string PkgVerRel,
        std::string PkgStatus, std::string PkdDepBrokenName, std::string PkgDepConstrain,
        std::string PkgDepConstVersion, std::string PkgDepCandidate) {

    DudfPackageInfo PkgElement;

    PkgElement.Name = PkgName;
    PkgElement.VerRel = PkgVerRel;
    PkgElement.Status = PkgStatus;
    PkgElement.DepBrokenName = PkdDepBrokenName;
    PkgElement.DepConstrain = PkgDepConstrain;
    PkgElement.DepConstVersion = PkgDepConstVersion;
    PkgElement.DepCandidate = PkgDepCandidate;

    PtrPkgInfo.push_back(PkgElement);

}
/*}}}*/

